@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2009-2022 Vadim Mikhailov
+ * Copyright (c) 2009-2024 Vadim Mikhailov
  *
  * Utility to turn USB port power on/off
  * for USB hubs that support per-port power switching.
@@ -29,7 +29,7 @@
 #include <unistd.h>
 #endif
 
-#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(_WIN32)
+#if defined(__APPLE__) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(_WIN32)
 #include <libusb.h>
 #else
 #include <libusb-1.0/libusb.h>
@@ -47,7 +47,7 @@ int snprintf(char * __restrict __str, size_t __size, const char * __restrict __f
 #include <time.h>   /* for nanosleep */
 #endif
 
-#ifdef __gnu_linux__
+#if defined(__gnu_linux__) || defined(__linux__)
 #include <fcntl.h> /* for open() / O_WRONLY */
 #endif
 
@@ -80,6 +80,7 @@ void sleep_ms(int milliseconds)
 #define POWER_ON                 1
 #define POWER_CYCLE              2
 #define POWER_TOGGLE             3
+#define POWER_FLASH              4
 
 #define MAX_HUB_CHAIN            8  /* Per USB 3.0 spec max hub chain is 7 */
 
@@ -226,14 +227,14 @@ static int opt_exact  = 0;  /* exact location match - disable USB3 duality handl
 static int opt_reset  = 0;  /* reset hub after operation(s) */
 static int opt_force  = 0;  /* force operation even on unsupported hubs */
 static int opt_nodesc = 0;  /* skip querying device description */
-#ifdef __gnu_linux__
+#if defined(__gnu_linux__) || defined(__linux__)
 static int opt_nosysfs = 0; /* don't use the Linux sysfs port disable interface, even if available */
 #endif
 
 
 static const char short_options[] =
     "l:L:n:a:p:d:r:w:s:hvefRN"
-#ifdef __gnu_linux__
+#if defined(__gnu_linux__) || defined(__linux__)
     "S"
 #endif
 ;
@@ -251,7 +252,7 @@ static const struct option long_options[] = {
     { "exact",    no_argument,       NULL, 'e' },
     { "force",    no_argument,       NULL, 'f' },
     { "nodesc",   no_argument,       NULL, 'N' },
-#ifdef __gnu_linux__
+#if defined(__gnu_linux__) || defined(__linux__)
     { "nosysfs",  no_argument,       NULL, 'S' },
 #endif
     { "reset",    no_argument,       NULL, 'R' },
@@ -261,7 +262,7 @@ static const struct option long_options[] = {
 };
 
 
-static int print_usage()
+static int print_usage(void)
 {
     printf(
         "uhubctl: utility to control USB port power for smart hubs.\n"
@@ -269,18 +270,18 @@ static int print_usage()
         "Without options, show status for all smart hubs.\n"
         "\n"
         "Options [defaults in brackets]:\n"
-        "--action,   -a - action to off/on/cycle/toggle (0/1/2/3) for affected ports.\n"
+        "--action,   -a - action to off/on/cycle/toggle/flash (0/1/2/3/4) for affected ports.\n"
         "--ports,    -p - ports to operate on    [all hub ports].\n"
         "--location, -l - limit hub by location  [all smart hubs].\n"
         "--level     -L - limit hub by location level (e.g. a-b.c is level 3).\n"
         "--vendor,   -n - limit hub by vendor id [%s] (partial ok).\n"
         "--search,   -s - limit hub by attached device description.\n"
-        "--delay,    -d - delay for cycle action [%g sec].\n"
+        "--delay,    -d - delay for cycle/flash action [%g sec].\n"
         "--repeat,   -r - repeat power off count [%d] (some devices need it to turn off).\n"
         "--exact,    -e - exact location (no USB3 duality handling).\n"
         "--force,    -f - force operation even on unsupported hubs.\n"
         "--nodesc,   -N - do not query device description (helpful for unresponsive devices).\n"
-#ifdef __gnu_linux__
+#if defined(__gnu_linux__) || defined(__linux__)
         "--nosysfs,  -S - do not use the Linux sysfs port disable interface.\n"
 #endif
         "--reset,    -R - reset hub after each power-on action, causing all devices to reassociate.\n"
@@ -404,7 +405,7 @@ static int get_hub_info(struct libusb_device *dev, struct hub_info *info)
         return rc;
     if (desc.bDeviceClass != LIBUSB_CLASS_HUB)
         return LIBUSB_ERROR_INVALID_PARAM;
-    int bcd_usb = libusb_le16_to_cpu(desc.bcdUSB);
+    int bcd_usb = desc.bcdUSB;
     int desc_type = bcd_usb >= USB_SS_BCD ? LIBUSB_DT_SUPERSPEED_HUB
                                           : LIBUSB_DT_HUB;
     rc = libusb_open(dev, &devh);
@@ -427,8 +428,8 @@ static int get_hub_info(struct libusb_device *dev, struct hub_info *info)
             snprintf(
                 info->vendor, sizeof(info->vendor),
                 "%04x:%04x",
-                libusb_le16_to_cpu(desc.idVendor),
-                libusb_le16_to_cpu(desc.idProduct)
+                desc.idVendor,
+                desc.idProduct
             );
 
             /* Convert bus and ports array into USB location string */
@@ -490,6 +491,32 @@ static int get_hub_info(struct libusb_device *dev, struct hub_info *info)
                 lpsm = HUB_CHAR_INDV_PORT_LPSM;
             }
             info->lpsm = lpsm;
+
+            /* Raspberry Pi 5 hack */
+
+            /* TODO: make this hack more reliable by querying Raspberry Pi model */
+
+            if (strlen(info->container_id)==0 &&
+                info->lpsm==HUB_CHAR_INDV_PORT_LPSM &&
+                info->pn_len==0)
+            {
+                /* USB2 */
+                if (strcasecmp(info->vendor, "1d6b:0002")==0 &&
+                    info->nports==2 &&
+                    !info->super_speed &&
+                    (info->bus==1 || info->bus==3))
+                {
+                    strcpy(info->container_id, "Raspberry Pi 5 Fake Container Id");
+                }
+                /* USB3 */
+                if (strcasecmp(info->vendor, "1d6b:0003")==0 &&
+                    info->nports==1 &&
+                    info->super_speed &&
+                    (info->bus==2 || info->bus==4))
+                {
+                    strcpy(info->container_id, "Raspberry Pi 5 Fake Container Id");
+                }
+            }
             rc = 0;
         } else {
             rc = len;
@@ -524,11 +551,11 @@ static int get_port_status(struct libusb_device_handle *devh, int port)
     if (rc < 0) {
         return rc;
     }
-    return ust.wPortStatus;
+    return libusb_le16_to_cpu(ust.wPortStatus);
 }
 
 
-#ifdef __gnu_linux__
+#if defined(__gnu_linux__) || defined(__linux__)
 /*
  * Try to use the Linux sysfs interface to power a port off/on.
  * Returns 0 on success.
@@ -548,10 +575,17 @@ static int set_port_status_linux(struct libusb_device_handle *devh, struct hub_i
      * The "disable" sysfs interface is available only starting with kernel version 6.0.
      * For earlier kernel versions the open() call will fail and we fall back to using libusb.
      */
-    snprintf(disable_path, PATH_MAX,
-        "/sys/bus/usb/devices/%s:%d.0/%s-port%i/disable",
-        hub->location, configuration, hub->location, port
-    );
+    if (hub->pn_len == 0) {
+      snprintf(disable_path, PATH_MAX,
+          "/sys/bus/usb/devices/%s-0:%d.0/usb%s-port%i/disable",
+          hub->location, configuration, hub->location, port
+      );
+    } else {
+      snprintf(disable_path, PATH_MAX,
+          "/sys/bus/usb/devices/%s:%d.0/%s-port%i/disable",
+          hub->location, configuration, hub->location, port
+      );
+    }
 
     int disable_fd = open(disable_path, O_WRONLY);
     if (disable_fd >= 0) {
@@ -621,7 +655,7 @@ static int set_port_status_libusb(struct libusb_device_handle *devh, int port, i
 
 static int set_port_status(struct libusb_device_handle *devh, struct hub_info *hub, int port, int on)
 {
-#ifdef __gnu_linux__
+#if defined(__gnu_linux__) || defined(__linux__)
     if (!opt_nosysfs) {
         if (set_port_status_linux(devh, hub, port, on) == 0) {
             return 0;
@@ -662,8 +696,8 @@ static int get_device_description(struct libusb_device * dev, struct descriptor_
     if (rc)
         return rc;
     bzero(ds, sizeof(*ds));
-    id_vendor  = libusb_le16_to_cpu(desc.idVendor);
-    id_product = libusb_le16_to_cpu(desc.idProduct);
+    id_vendor  = desc.idVendor;
+    id_product = desc.idProduct;
     rc = libusb_open(dev, &devh);
     if (rc == 0) {
         if (!opt_nodesc) {
@@ -685,6 +719,7 @@ static int get_device_description(struct libusb_device * dev, struct descriptor_
         }
         if (desc.bDeviceClass == LIBUSB_CLASS_HUB) {
             struct hub_info info;
+            bzero(&info, sizeof(info));
             rc = get_hub_info(dev, &info);
             if (rc == 0) {
                 const char * lpsm_type;
@@ -823,7 +858,7 @@ static int print_port_status(struct hub_info * hub, int portmask)
  *  In case of error returns negative error code.
  */
 
-static int usb_find_hubs()
+static int usb_find_hubs(void)
 {
     struct libusb_device *dev;
     int perm_ok = 1;
@@ -1011,7 +1046,7 @@ static int usb_find_hubs()
         }
     }
     if (perm_ok == 0 && hub_phys_count == 0) {
-#ifdef __gnu_linux__
+#if defined(__gnu_linux__) || defined(__linux__)
         if (geteuid() != 0) {
             fprintf(stderr,
                 "There were permission problems while accessing USB.\n"
@@ -1078,6 +1113,9 @@ int main(int argc, char *argv[])
             if (!strcasecmp(optarg, "toggle") || !strcasecmp(optarg, "3")) {
                 opt_action = POWER_TOGGLE;
             }
+            if (!strcasecmp(optarg, "flash") || !strcasecmp(optarg, "4")) {
+                opt_action = POWER_FLASH;
+            }
             break;
         case 'd':
             opt_delay = atof(optarg);
@@ -1091,7 +1129,7 @@ int main(int argc, char *argv[])
         case 'N':
             opt_nodesc = 1;
             break;
-#ifdef __gnu_linux__
+#if defined(__gnu_linux__) || defined(__linux__)
         case 'S':
             opt_nosysfs = 1;
             break;
@@ -1193,6 +1231,9 @@ int main(int argc, char *argv[])
                 /* will operate on these ports */
                 int ports = ((1 << hubs[i].nports) - 1) & opt_ports;
                 int should_be_on = k;
+                if (opt_action == POWER_FLASH) {
+                    should_be_on = !should_be_on;
+                }
 
                 int port;
                 for (port=1; port <= hubs[i].nports; port++) {
@@ -1232,7 +1273,7 @@ int main(int argc, char *argv[])
             }
             libusb_close(devh);
         }
-        if (k == 0 && opt_action == POWER_CYCLE)
+        if (k == 0 && (opt_action == POWER_CYCLE || opt_action == POWER_FLASH))
             sleep_ms((int)(opt_delay * 1000));
     }
     rc = 0;
